@@ -5,7 +5,8 @@
 import           Control.Monad.IO.Class       (liftIO)
 import           Data.XML.Types
 import           Test.HUnit                   hiding (Test)
-import           Test.Hspec.Monadic
+import           Test.Hspec
+import qualified Data.ByteString.Char8        as S
 import qualified Data.ByteString.Lazy.Char8   as L
 import qualified Text.XML.Unresolved          as D
 import qualified Text.XML.Stream.Parse        as P
@@ -19,12 +20,14 @@ import Control.Monad
 import qualified Data.Text as T
 import qualified Data.Set as Set
 import Control.Exception (toException)
-import Test.Hspec.HUnit ()
 
 import qualified Data.Conduit as C
+import qualified Data.Map as Map
+import Text.Blaze (toMarkup)
+import Text.Blaze.Renderer.String (renderMarkup)
 
 main :: IO ()
-main = hspecX $ do
+main = hspec $ do
     describe "XML parsing and rendering" $ do
         it "is idempotent to parse and render a document" documentParseRender
         it "has valid parser combinators" combinators
@@ -66,6 +69,18 @@ main = hspecX $ do
         it "works for resolvable entities" resolvedAllGood
         it "merges adjacent content nodes" resolvedMergeContent
         it "understands inline entity declarations" resolvedInline
+    describe "pretty" $ do
+        it "works" casePretty
+    describe "top level namespaces" $ do
+        it "works" caseTopLevelNamespace
+        it "works with prefix" caseTopLevelNamespacePrefix
+        it "handles conflicts" caseTLNConflict
+    describe "blaze-html instances" $ do
+        it "works" caseBlazeHtml
+    describe "attribute reordering" $ do
+        it "works" caseAttrReorder
+    describe "ordering attributes explicitly" $ do
+        it "works" caseOrderAttrs
 
 documentParseRender :: IO ()
 documentParseRender =
@@ -77,9 +92,9 @@ documentParseRender =
                    (Element "foo" [] [])
                    []
         , D.parseLBS_ def
-            "<?xml version=\"1.0\"?>\n<!DOCTYPE foo>\n<foo/>"
+            "<?xml version=\"1.0\"?><!DOCTYPE foo>\n<foo/>"
         , D.parseLBS_ def
-            "<?xml version=\"1.0\"?>\n<!DOCTYPE foo>\n<foo><nested>&ignore;</nested></foo>"
+            "<?xml version=\"1.0\"?><!DOCTYPE foo>\n<foo><nested>&ignore;</nested></foo>"
         , D.parseLBS_ def
             "<foo><![CDATA[this is some<CDATA content>]]></foo>"
         , D.parseLBS_ def
@@ -92,15 +107,13 @@ documentParseRender =
 
 documentParsePrettyRender :: IO ()
 documentParsePrettyRender =
-    L.unpack (D.renderLBS def { D.rsPretty = True } (D.parseLBS_ def $ doc True)) @?= L.unpack (doc False)
+    L.unpack (D.renderLBS def { D.rsPretty = True } (D.parseLBS_ def doc)) @?= L.unpack doc
   where
-    doc x = L.unlines
+    doc = L.unlines
         [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
         , "<foo>"
         , "    <?bar bar?>"
-        , if x
-            then "    text"
-            else "    text     "
+        , "    text"
         , "    <?bin bin?>"
         , "</foo>"
         ]
@@ -116,7 +129,7 @@ combinators = C.runResourceT $ P.parseLBS def input C.$$ do
             liftIO $ x @?= Just "combine <all> &content"
   where
     input = L.concat
-        [ "<?xml version='1.0'?>\n"
+        [ "<?xml version='1.0'?>"
         , "<!DOCTYPE foo []>\n"
         , "<hello world='true'>"
         , "<?this should be ignored?>"
@@ -137,7 +150,7 @@ testChoose = C.runResourceT $ P.parseLBS def input C.$$ do
         liftIO $ x @?= Just (2 :: Int)
   where
     input = L.concat
-        [ "<?xml version='1.0'?>\n"
+        [ "<?xml version='1.0'?>"
         , "<!DOCTYPE foo []>\n"
         , "<hello>"
         , "<success/>"
@@ -151,7 +164,7 @@ testMany = C.runResourceT $ P.parseLBS def input C.$$ do
         liftIO $ length x @?= 5
   where
     input = L.concat
-        [ "<?xml version='1.0'?>\n"
+        [ "<?xml version='1.0'?>"
         , "<!DOCTYPE foo []>\n"
         , "<hello>"
         , "<success/>"
@@ -170,7 +183,7 @@ testOrE = C.runResourceT $ P.parseLBS def input C.$$ do
         liftIO $ x @?= Just (2 :: Int)
   where
     input = L.concat
-        [ "<?xml version='1.0'?>\n"
+        [ "<?xml version='1.0'?>"
         , "<!DOCTYPE foo []>\n"
         , "<hello>"
         , "<success/>"
@@ -295,7 +308,7 @@ resolvedAllGood =
 
 resolvedMergeContent =
     Res.documentRoot (Res.parseLBS_ def xml) @=?
-    Res.Element "foo" [] [Res.NodeContent "bar&baz"]
+    Res.Element "foo" Map.empty [Res.NodeContent "bar&baz"]
   where
     xml = "<foo>bar&amp;baz</foo>"
 
@@ -306,9 +319,9 @@ parseIgnoreBOM = do
 
 stripDuplicateAttributes :: Assertion
 stripDuplicateAttributes = do
-    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<foo bar=\"baz\"/>" @=?
+    "<?xml version=\"1.0\" encoding=\"UTF-8\"?><foo bar=\"baz\"/>" @=?
         D.renderLBS def (Document (Prologue [] Nothing []) (Element "foo" [("bar", [ContentText "baz"]), ("bar", [ContentText "bin"])] []) [])
-    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<foo x:bar=\"baz\" xmlns:x=\"namespace\"/>" @=?
+    "<?xml version=\"1.0\" encoding=\"UTF-8\"?><foo x:bar=\"baz\" xmlns:x=\"namespace\"/>" @=?
         D.renderLBS def (Document (Prologue [] Nothing []) (Element "foo"
             [ ("x:bar", [ContentText "baz"])
             , (Name "bar" (Just "namespace") (Just "x"), [ContentText "bin"])
@@ -316,13 +329,170 @@ stripDuplicateAttributes = do
 
 testRenderComments :: Assertion
 testRenderComments =do
-    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<foo><!--comment--></foo>"
+    "<?xml version=\"1.0\" encoding=\"UTF-8\"?><foo><!--comment--></foo>"
         @=? D.renderLBS def (Document (Prologue [] Nothing [])
             (Element "foo" [] [NodeComment "comment"]) [])
 
 resolvedInline :: Assertion
 resolvedInline = do
     Res.Document _ root _ <- return $ Res.parseLBS_ Res.def "<!DOCTYPE foo [<!ENTITY bar \"baz\">]><foo>&bar;</foo>"
-    root @?= Res.Element "foo" [] [Res.NodeContent "baz"]
+    root @?= Res.Element "foo" Map.empty [Res.NodeContent "baz"]
     Res.Document _ root2 _ <- return $ Res.parseLBS_ Res.def "<!DOCTYPE foo [<!ENTITY bar \"baz\">]><foo bar='&bar;'/>"
-    root2 @?= Res.Element "foo" [("bar", "baz")] []
+    root2 @?= Res.Element "foo" (Map.singleton "bar" "baz") []
+
+casePretty :: Assertion
+casePretty = do
+    let pretty = S.unlines
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<!DOCTYPE foo>"
+            , "<foo bar=\"bar\" baz=\"baz\">"
+            , "    <foo"
+            , "      bar=\"bar\""
+            , "      baz=\"baz\""
+            , "      bin=\"bin\">"
+            , "        Hello World"
+            , "    </foo>"
+            , "    <foo/>"
+            , "    <?foo bar?>"
+            , "    <!-- foo bar baz bin -->"
+            , "    <bar>"
+            , "        bar content"
+            , "    </bar>"
+            , "</foo>"
+            ]
+        doctype = Res.Doctype "foo" Nothing
+        doc = Res.Document (Res.Prologue [] (Just doctype) []) root []
+        root = Res.Element "foo" (Map.fromList [("bar", "bar"), ("baz", "baz")])
+                [ Res.NodeElement $ Res.Element "foo" (Map.fromList [("bar", "bar"), ("baz", "baz"), ("bin", "bin")])
+                    [ Res.NodeContent "  Hello World\n\n"
+                    , Res.NodeContent "  "
+                    ]
+                , Res.NodeElement $ Res.Element "foo" Map.empty []
+                , Res.NodeInstruction $ Res.Instruction "foo" "bar"
+                , Res.NodeComment "foo bar\n\r\nbaz    \tbin "
+                , Res.NodeElement $ Res.Element "bar" Map.empty [Res.NodeContent "bar content"]
+                ]
+    pretty @=? S.concat (L.toChunks $ Res.renderLBS def { D.rsPretty = True } doc)
+
+caseTopLevelNamespace :: Assertion
+caseTopLevelNamespace = do
+    let lbs = S.concat
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<foo xmlns:bar=\"baz\">"
+            , "<subfoo bar:bin=\"\"/>"
+            , "</foo>"
+            ]
+        rs = def { D.rsNamespaces = [("bar", "baz")] }
+        doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "foo" Map.empty
+                    [ Res.NodeElement
+                        $ Res.Element "subfoo" (Map.singleton "{baz}bin" "") []
+                    ])
+                []
+    lbs @=? S.concat (L.toChunks $ Res.renderLBS rs doc)
+
+caseTopLevelNamespacePrefix :: Assertion
+caseTopLevelNamespacePrefix = do
+    let lbs = S.concat
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<foo xmlns:bar=\"baz\">"
+            , "<subfoo bar:bin=\"\"/>"
+            , "</foo>"
+            ]
+        rs = def { D.rsNamespaces = [("bar", "baz")] }
+        doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "foo" Map.empty
+                    [ Res.NodeElement
+                        $ Res.Element "subfoo" (Map.fromList [(Name "bin" (Just "baz") (Just "bar"), "")]) []
+                    ])
+                []
+    lbs @=? S.concat (L.toChunks $ Res.renderLBS rs doc)
+
+caseTLNConflict :: Assertion
+caseTLNConflict = do
+    let lbs = S.concat
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<foo xmlns:bar=\"something\" bar:x=\"y\">"
+            , "<subfoo xmlns:bar_=\"baz\" bar_:bin=\"\"/>"
+            , "</foo>"
+            ]
+        rs = def { D.rsNamespaces = [("bar", "baz")] }
+        doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "foo" (Map.fromList [(Name "x" (Just "something") (Just "bar"), "y")])
+                    [ Res.NodeElement
+                        $ Res.Element "subfoo" (Map.fromList [(Name "bin" (Just "baz") (Just "bar"), "")]) []
+                    ])
+                []
+    lbs @=? S.concat (L.toChunks $ Res.renderLBS rs doc)
+
+caseBlazeHtml :: Assertion
+caseBlazeHtml =
+    expected @=? str
+  where
+    str = renderMarkup $ toMarkup $ Res.Document (Res.Prologue [] Nothing []) root []
+    root :: Res.Element
+    root = Res.Element "html" Map.empty
+        [ Res.NodeElement $ Res.Element "head" Map.empty
+            [ Res.NodeElement $ Res.Element "title" Map.empty [Res.NodeContent "Test"]
+            , Res.NodeElement $ Res.Element "script" Map.empty
+                [Res.NodeContent "if (5 < 6 || 8 > 9) alert('Hello World!');"]
+            , Res.NodeElement $ Res.Element "{http://www.snoyman.com/xml2html}ie-cond" (Map.singleton "cond" "lt IE 7")
+                [Res.NodeElement $ Res.Element "link" (Map.singleton "href" "ie6.css") []]
+            , Res.NodeElement $ Res.Element "style" Map.empty
+                [Res.NodeContent "body > h1 { color: red }"]
+            ]
+        , Res.NodeElement $ Res.Element "body" Map.empty
+            [ Res.NodeElement $ Res.Element "h1" Map.empty [Res.NodeContent "Hello World!"]
+            ]
+        ]
+    expected :: String
+    expected = concat
+        [ "<!DOCTYPE HTML>\n"
+        , "<html><head><title>Test</title><script>if (5 < 6 || 8 > 9) alert('Hello World!');</script>"
+        , "<!--[if lt IE 7]><link href=\"ie6.css\" /><![endif]-->"
+        , "<style>body > h1 { color: red }</style>"
+        , "</head>"
+        , "<body><h1>Hello World!</h1></body></html>"
+        ]
+
+caseAttrReorder :: Assertion
+caseAttrReorder = do
+    let lbs = S.concat
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<foo c=\"c\" b=\"b\" a=\"a\">"
+            , "<bar a=\"a\" b=\"b\" c=\"c\"/>"
+            , "</foo>"
+            ]
+        rs = def { Res.rsAttrOrder = \name m ->
+                        case name of
+                            "foo" -> reverse $ Map.toAscList m
+                            _ -> Map.toAscList m
+                 }
+        attrs = Map.fromList [("a", "a"), ("b", "b"), ("c", "c")]
+        doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "foo" attrs
+                    [ Res.NodeElement
+                        $ Res.Element "bar" attrs []
+                    ])
+                []
+    lbs @=? S.concat (L.toChunks $ Res.renderLBS rs doc)
+
+caseOrderAttrs :: Assertion
+caseOrderAttrs = do
+    let lbs = S.concat
+            [ "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
+            , "<foo c=\"c\" b=\"b\" a=\"a\">"
+            , "<bar a=\"a\" b=\"b\" c=\"c\"/>"
+            , "</foo>"
+            ]
+        rs = def { Res.rsAttrOrder = Res.orderAttrs
+                     [("foo", ["c", "b"])]
+                 }
+        attrs = Map.fromList [("a", "a"), ("b", "b"), ("c", "c")]
+        doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "foo" attrs
+                    [ Res.NodeElement
+                        $ Res.Element "bar" attrs []
+                    ])
+                []
+    lbs @=? S.concat (L.toChunks $ Res.renderLBS rs doc)
